@@ -18,6 +18,8 @@
 #include "php_excimer.h"
 #include "excimer_log.h"
 
+static const char excimer_log_truncated_name[] = "excimer_truncated";
+static const char excimer_log_fake_filename[] = "excimer fake file";
 
 static uint32_t excimer_log_find_or_add_frame(excimer_log *log,
 		zend_execute_data *execute_data, zend_long depth);
@@ -30,7 +32,7 @@ static uint32_t excimer_log_find_or_add_frame(excimer_log *log,
 static inline HashTable *excimer_log_new_array(uint32_t nSize)
 {
 	HashTable *ht = emalloc(sizeof(HashTable));
-	zend_hash_init(ht, 0, NULL, ZVAL_PTR_DTOR, 0);
+	zend_hash_init(ht, nSize, NULL, ZVAL_PTR_DTOR, 0);
 	return ht;
 }
 #endif
@@ -75,6 +77,17 @@ static void excimer_log_smart_str_append_printf(smart_str *dest, const char *for
 #define excimer_log_known_string(index) CG(known_strings)[index]
 #endif
 
+#if PHP_VERSION_ID >= 80100
+#define excimer_log_add_assoc_array add_assoc_array
+#else
+static inline void excimer_log_add_assoc_array(zval *dest, const char *key, HashTable *arr)
+{
+	zval z_tmp;
+	ZVAL_ARR(&z_tmp, arr);
+	add_assoc_zval(dest, key, &z_tmp);
+}
+#endif
+
 /* }}} */
 
 void excimer_log_init(excimer_log *log)
@@ -121,6 +134,7 @@ void excimer_log_copy_options(excimer_log *dest, excimer_log  *src)
 {
 	dest->max_depth = src->max_depth;
 	dest->epoch = src->epoch;
+	dest->period = src->period;
 }
 
 void excimer_log_add(excimer_log *log, zend_execute_data *execute_data,
@@ -138,19 +152,50 @@ void excimer_log_add(excimer_log *log, zend_execute_data *execute_data,
 	entry->timestamp = timestamp;
 }
 
+static uint32_t excimer_log_get_truncation_marker(excimer_log *log) {
+	zval* zp_index;
+	zval z_new_index;
+	excimer_log_frame *p_frame;
+	
+	zp_index = zend_hash_str_find(log->reverse_frames,
+		excimer_log_truncated_name, sizeof(excimer_log_truncated_name) - 1);
+	if (zp_index) {
+		return excimer_safe_uint32(Z_LVAL_P(zp_index));
+	}
+
+	ZVAL_LONG(&z_new_index, log->frames_size);
+	zend_hash_str_add(log->reverse_frames,
+		excimer_log_truncated_name, sizeof(excimer_log_truncated_name) - 1,
+		&z_new_index);
+	log->frames = safe_erealloc(log->frames, log->frames_size + 1,
+		sizeof(excimer_log_frame), 0);
+	p_frame = &log->frames[log->frames_size++];
+
+	p_frame->filename = zend_string_init(excimer_log_fake_filename,
+		sizeof(excimer_log_fake_filename) - 1, 0);
+	p_frame->lineno = 1;
+	p_frame->closure_line = 0;
+	p_frame->class_name = NULL;
+	p_frame->function_name = zend_string_init(excimer_log_truncated_name,
+		sizeof(excimer_log_truncated_name) - 1, 0);
+	p_frame->prev_index = 0;
+
+	return excimer_safe_uint32(Z_LVAL(z_new_index));
+}
+
 static uint32_t excimer_log_find_or_add_frame(excimer_log *log,
 	zend_execute_data *execute_data, zend_long depth)
 {
 	uint32_t prev_index;
 	if (!execute_data) {
 		return 0;
-	} else if (execute_data->prev_execute_data
-		&& (!log->max_depth || depth < log->max_depth))
-	{
+	} else if (!execute_data->prev_execute_data) {
+		prev_index = 0;
+	} else if (log->max_depth && depth >= log->max_depth) {
+		prev_index = excimer_log_get_truncation_marker(log);
+	} else {
 		prev_index = excimer_log_find_or_add_frame(log,
 			execute_data->prev_execute_data, depth + 1);
-	} else {
-		prev_index = 0;
 	}
 	if (!execute_data->func
 		|| !ZEND_USER_CODE(execute_data->func->common.type))
@@ -258,40 +303,68 @@ static void excimer_log_append_no_spaces(smart_str *dest, zend_string *src)
 	ZSTR_LEN(dest->s) = new_len;
 }
 
+static void excimer_log_append_frame_name(smart_str *ss, excimer_log_frame *frame) {
+	if (frame->closure_line != 0) {
+		/* Annotate anonymous functions with their source location.
+		 * Example: {closure:/path/to/file.php(123)}
+		 */
+		smart_str_appends(ss, "{closure:");
+		excimer_log_append_no_spaces(ss, frame->filename);
+		excimer_log_smart_str_append_printf(ss, "(%d)}", frame->closure_line);
+	} else if (frame->function_name == NULL) {
+		/* For file-scope code, use the file name */
+		excimer_log_append_no_spaces(ss, frame->filename);
+	} else {
+		if (frame->class_name) {
+			excimer_log_append_no_spaces(ss, frame->class_name);
+			smart_str_appends(ss, "::");
+		}
+		excimer_log_append_no_spaces(ss, frame->function_name);
+	}
+}
+
 zend_string *excimer_log_format_collapsed(excimer_log *log)
 {
 	zend_long entry_index;
 	zend_long frame_index;
 	zval *zp_count;
-	smart_str buf = {NULL};
-	HashTable ht_storage;
-	HashTable *ht;
+	zval z_count;
+	smart_str ss_out = {NULL};
+	HashTable frame_counts_storage, lines_storage;
+	HashTable *ht_frame_counts, *ht_lines;
+
+	ht_frame_counts = &frame_counts_storage;
+	memset(ht_frame_counts, 0, sizeof(HashTable));
+	zend_hash_init(ht_frame_counts, 0, NULL, NULL, 0);
+
+	ht_lines = &lines_storage;
+	memset(ht_lines, 0, sizeof(HashTable));
+	zend_hash_init(ht_lines, 0, NULL, NULL, 0);
 
-	ht = &ht_storage;
-	memset(ht, 0, sizeof(HashTable));
-	zend_hash_init(ht, 0, NULL, NULL, 0);
 	excimer_log_frame ** frame_ptrs = NULL;
 	size_t frames_capacity = 0;
+	zend_string *str_line;
 
 	/* Collate frame counts */
 	for (entry_index = 0; entry_index < log->entries_size; entry_index++) {
 		excimer_log_entry *entry = excimer_log_get_entry(log, entry_index);
-		zp_count = zend_hash_index_find(ht, entry->frame_index);
+		zp_count = zend_hash_index_find(ht_frame_counts, entry->frame_index);
 		if (!zp_count) {
-			zval z_count;
 			ZVAL_LONG(&z_count, 0);
-			zp_count = zend_hash_index_add(ht, entry->frame_index, &z_count);
+			zp_count = zend_hash_index_add(ht_frame_counts, entry->frame_index, &z_count);
 		}
 
 		Z_LVAL_P(zp_count) += entry->event_count;
 	}
 
-	/* Format traces */
-	ZEND_HASH_FOREACH_NUM_KEY_VAL(ht, frame_index, zp_count) {
+	/* Format traces, and deduplicate frames that differ only in hidden line numbers */
+	ZEND_HASH_FOREACH_NUM_KEY_VAL(ht_frame_counts, frame_index, zp_count) {
 		zend_long current_frame_index = frame_index;
 		zend_long num_frames = 0;
 		excimer_log_frame *frame;
 		zend_long i;
+		int line_start = 1; /* TODO use bool when PHP 7.4 support is dropped */
+		smart_str ss_line = {NULL};
 
 		/* Build the array of frame pointers */
 		while (current_frame_index) {
@@ -312,36 +385,179 @@ zend_string *excimer_log_format_collapsed(excimer_log *log)
 		for (i = num_frames - 1; i >= 0; i--) {
 			frame = frame_ptrs[i];
 
-			if (excimer_log_smart_str_get_len(&buf) != 0) {
-				smart_str_appends(&buf, ";");
-			}
-
-			if (frame->closure_line != 0) {
-				/* Annotate anonymous functions with their source location.
-				 * Example: {closure:/path/to/file.php(123)}
-				 */
-				smart_str_appends(&buf, "{closure:");
-				excimer_log_append_no_spaces(&buf, frame->filename);
-				excimer_log_smart_str_append_printf(&buf, "(%d)}", frame->closure_line);
-			} else if (frame->function_name == NULL) {
-				/* For file-scope code, use the file name */
-				excimer_log_append_no_spaces(&buf, frame->filename);
+			if (line_start) {
+				line_start = 0;
 			} else {
-				if (frame->class_name) {
-					excimer_log_append_no_spaces(&buf, frame->class_name);
-					smart_str_appends(&buf, "::");
-				}
-				excimer_log_append_no_spaces(&buf, frame->function_name);
+				smart_str_appends(&ss_line, ";");
 			}
+			excimer_log_append_frame_name(&ss_line, frame);
 		}
 
-		/* Append count and line break */
-		excimer_log_smart_str_append_printf(&buf, " %ld\n", Z_LVAL_P(zp_count));
+		/* ht_lines[ss_line] += zp_count */
+		str_line = excimer_log_smart_str_extract(&ss_line);
+		zval *zp_line_count = zend_hash_find(ht_lines, str_line);
+		if (!zp_line_count) {
+			ZVAL_LONG(&z_count, 0);
+			zp_line_count = zend_hash_add(ht_lines, str_line, &z_count);
+		}
+		Z_LVAL_P(zp_line_count) += Z_LVAL_P(zp_count);
 	}
 	ZEND_HASH_FOREACH_END();
-	zend_hash_destroy(ht);
+
+	/* Concatenate lines */
+	ZEND_HASH_FOREACH_STR_KEY_VAL(ht_lines, str_line, zp_count) {
+		smart_str_append(&ss_out, str_line);
+		excimer_log_smart_str_append_printf(&ss_out, " %ld\n", Z_LVAL_P(zp_count));
+	}
+	ZEND_HASH_FOREACH_END();
+
+	zend_hash_destroy(ht_frame_counts);
+	zend_hash_destroy(ht_lines);
 	efree(frame_ptrs);
-	return excimer_log_smart_str_extract(&buf);
+	return excimer_log_smart_str_extract(&ss_out);
+}
+
+static HashTable *excimer_log_frame_to_speedscope_array(excimer_log_frame *frame) {
+	HashTable *ht_func = excimer_log_new_array(0);
+	zval tmp;
+	smart_str ss_name = {NULL};
+	
+	excimer_log_append_frame_name(&ss_name, frame);
+	ZVAL_STR(&tmp, excimer_log_smart_str_extract(&ss_name));
+	zend_hash_str_add(ht_func, "name", sizeof("name")-1, &tmp);
+
+	if (frame->filename) {
+		ZVAL_STR_COPY(&tmp, frame->filename);
+		zend_hash_add_new(ht_func, excimer_log_known_string(ZEND_STR_FILE), &tmp);
+		/* Don't include the line number since it causes speedscope to split functions */
+	}
+	return ht_func;
+}
+
+static zend_string *excimer_log_get_speedscope_frame_key(excimer_log_frame *frame) {
+	smart_str ss = {NULL};
+	
+	excimer_log_append_frame_name(&ss, frame);
+	smart_str_appendc(&ss, '\0');
+	smart_str_append(&ss, frame->filename);
+	return excimer_log_smart_str_extract(&ss);
+}
+
+static uint32_t excimer_log_count_frames(excimer_log *log, uint32_t frame_index) {
+	uint32_t n = 0;
+	while (frame_index) {
+		n++;
+		frame_index = log->frames[frame_index].prev_index;
+	}
+	return n;
+}
+
+void excimer_log_get_speedscope_data(excimer_log *log, zval *zp_data) {
+	array_init(zp_data);
+	add_assoc_string(zp_data, "$schema", "https://www.speedscope.app/file-format-schema.json");
+	add_assoc_string(zp_data, "exporter", "Excimer");
+
+	HashTable *ht_frames = excimer_log_new_array(0);
+	HashTable *ht_indexes_by_key = excimer_log_new_array(0);
+	zend_long *lp_frame_indexes = ecalloc(log->frames_size, sizeof(zend_long));
+	zend_long i;
+	zval *zp_frame_index;
+	zend_string *str_key;
+	zval z_tmp, *zp_tmp;
+
+	/* Build the frames array */
+	for (i = 1; i < log->frames_size; i++) {
+		zend_long index;
+		excimer_log_frame *frame = &log->frames[i];
+		str_key = excimer_log_get_speedscope_frame_key(frame);
+		zp_frame_index = zend_hash_find(ht_indexes_by_key, str_key);
+		if (!zp_frame_index) {
+			/* Add the frame to ht_frames */
+			index = zend_hash_num_elements(ht_frames);
+			ZVAL_ARR(&z_tmp, excimer_log_frame_to_speedscope_array(frame));
+			zend_hash_next_index_insert_new(ht_frames, &z_tmp);
+			/* Add the frame index to ht_indexes_by_key */
+			ZVAL_LONG(&z_tmp, index);
+			zp_frame_index = zend_hash_add_new(ht_indexes_by_key, str_key, &z_tmp);
+		}
+		lp_frame_indexes[i] = Z_LVAL_P(zp_frame_index);
+	}
+
+	/* zp_data["shared"] = ["frames" => ht_frames] */
+	zval z_shared;
+	array_init(&z_shared);
+	excimer_log_add_assoc_array(&z_shared, "frames", ht_frames);
+	add_assoc_zval(zp_data, "shared", &z_shared);
+
+	/* Build the samples and weights arrays */
+	HashTable *ht_samples = excimer_log_new_array(log->entries_size);
+	HashTable *ht_weights = excimer_log_new_array(log->entries_size);
+	uint64_t first_timestamp = 0;
+	uint64_t last_timestamp = 0;
+	for (i = 0; i < log->entries_size; i++) {
+		excimer_log_entry *entry = &log->entries[i];
+		uint32_t frame_index = entry->frame_index;
+
+		if (i == 0) {
+			first_timestamp = entry->timestamp;
+		}
+		last_timestamp = entry->timestamp;
+
+		uint32_t num_frames = excimer_log_count_frames(log, frame_index);
+		uint32_t j;
+
+		/* Create the array with ZEND_HASH_FILL_PACKED. This is just a fast way
+		 * to get it into the right state, with num_frames elements. */
+		HashTable *ht_stack = excimer_log_new_array(num_frames);
+		zend_hash_extend(ht_stack, num_frames, 1);
+		ZEND_HASH_FILL_PACKED(ht_stack) {
+#if PHP_VERSION_ID < 70400
+			zval new_val;
+
+			ZVAL_LONG(&new_val, 0);
+			for (j = 0; j < num_frames; j++) {
+				ZEND_HASH_FILL_ADD(&new_val);
+			}
+#else
+			for (j = 0; j < num_frames; j++) {
+				ZEND_HASH_FILL_SET_LONG(0);
+				ZEND_HASH_FILL_NEXT();
+			}
+#endif
+		} ZEND_HASH_FILL_END();
+
+		/* Write the values in reverse order */
+		ZEND_HASH_REVERSE_FOREACH_VAL(ht_stack, zp_tmp) {
+			ZVAL_LONG(zp_tmp, lp_frame_indexes[frame_index]);
+			frame_index = log->frames[frame_index].prev_index;
+		}
+		ZEND_HASH_FOREACH_END();
+
+		ZVAL_ARR(&z_tmp, ht_stack);
+		zend_hash_next_index_insert_new(ht_samples, &z_tmp);
+
+		ZVAL_LONG(&z_tmp, entry->event_count * log->period);
+		zend_hash_next_index_insert_new(ht_weights, &z_tmp);
+	}
+
+	/* Build the profile array */
+	zval z_profile;
+	array_init(&z_profile);
+	add_assoc_string(&z_profile, "type", "sampled");
+	add_assoc_string(&z_profile, "name", "");
+	add_assoc_string(&z_profile, "unit", "nanoseconds");
+	add_assoc_long(&z_profile, "startValue", 0);
+	add_assoc_long(&z_profile, "endValue", last_timestamp - first_timestamp);
+	excimer_log_add_assoc_array(&z_profile, "samples", ht_samples);
+	excimer_log_add_assoc_array(&z_profile, "weights", ht_weights);
+
+	/* zp_data["profiles"] = [profile] */
+	zval z_profiles;
+	array_init(&z_profiles);
+	add_next_index_zval(&z_profiles, &z_profile);
+	add_assoc_zval(zp_data, "profiles", &z_profiles);
+
+	efree(lp_frame_indexes);
 }
 
 HashTable *excimer_log_frame_to_array(excimer_log_frame *frame) {
